/*############################################################################*/
/*#                                                                          #*/
/*#  Ambisonic C++ Library                                                   #*/
/*#  Copyright © 2007 Aristotel Digenis                                      #*/
/*#                                                                          #*/
/*#  Filename:      Ambisonics.h                                             #*/
/*#  Version:       0.1                                                      #*/
/*#  Date:          19/05/2007                                               #*/
/*#  Author(s):     Aristotel Digenis                                        #*/
/*#  Licence:       MIT                                                      #*/
/*#                                                                          #*/
/*############################################################################*/


#ifndef SPATIALAUDIO_AMBISONICS_H
#define SPATIALAUDIO_AMBISONICS_H

#include "AmbisonicCommons.h"
#include "AmbisonicBase.h"
#include "BFormat.h"
#include "AmbisonicSource.h"
#include "AmbisonicSpeaker.h"
#include "AmbisonicMicrophone.h"
#include "AmbisonicEncoder.h"
#include "AmbisonicEncoderDist.h"
#include "AmbisonicDecoder.h"
#include "AmbisonicProcessor.h"
#include "AmbisonicRotator.h"
#include "AmbisonicBinauralizer.h"
#include "AmbisonicZoomer.h"
#include "AmbisonicShelfFilters.h"
#include "AmbisonicAllRAD.h"

#endif // SPATIALAUDIO_AMBISONICS_H
